@@ -72,6 +72,9 @@ struct connman_technology {
 	guint pending_timeout;
 
 	GSList *scan_pending;
+
+	connman_bool_t hardblocked;
+	connman_bool_t dbus_registered;
 };
 
 static GSList *driver_list = NULL;
@@ -244,6 +247,21 @@ void connman_technology_regdom_notify(struct connman_technology *technology,
 	technology->regdom = g_strdup(alpha2);
 }
 
+static int set_regdom_by_device(struct connman_technology *technology,
+							const char *alpha2)
+{
+	GSList *list;
+
+	for (list = technology->device_list; list; list = list->next) {
+		struct connman_device *device = list->data;
+
+		if (connman_device_set_regdom(device, alpha2) != 0)
+			return -ENOTSUP;
+	}
+
+	return 0;
+}
+
 int connman_technology_set_regdom(const char *alpha2)
 {
 	GSList *list;
@@ -251,11 +269,14 @@ int connman_technology_set_regdom(const char *alpha2)
 	for (list = technology_list; list; list = list->next) {
 		struct connman_technology *technology = list->data;
 
-		if (technology->driver == NULL)
-			continue;
+		if (set_regdom_by_device(technology, alpha2) != 0) {
+			if (technology->driver == NULL)
+				continue;
 
-		if (technology->driver->set_regdom)
-			technology->driver->set_regdom(technology, alpha2);
+			if (technology->driver->set_regdom != NULL)
+				technology->driver->set_regdom(technology,
+								alpha2);
+		}
 	}
 
 	return 0;
@@ -292,7 +313,7 @@ static const char *get_name(enum connman_service_type type)
 	return NULL;
 }
 
-static void save_state(struct connman_technology *technology)
+static void technology_save(struct connman_technology *technology)
 {
 	GKeyFile *keyfile;
 	gchar *identifier;
@@ -310,6 +331,16 @@ static void save_state(struct connman_technology *technology)
 	g_key_file_set_boolean(keyfile, identifier, "Enable",
 				technology->enable_persistent);
 
+	if (technology->tethering_ident != NULL)
+		g_key_file_set_string(keyfile, identifier,
+					"Tethering.Identifier",
+					technology->tethering_ident);
+
+	if (technology->tethering_passphrase != NULL)
+		g_key_file_set_string(keyfile, identifier,
+					"Tethering.Passphrase",
+					technology->tethering_passphrase);
+
 done:
 	g_free(identifier);
 
@@ -320,7 +351,7 @@ done:
 	return;
 }
 
-static void load_state(struct connman_technology *technology)
+static void technology_load(struct connman_technology *technology)
 {
 	GKeyFile *keyfile;
 	gchar *identifier;
@@ -353,9 +384,15 @@ static void load_state(struct connman_technology *technology)
 		else
 			technology->enable_persistent = FALSE;
 
-		save_state(technology);
+		technology_save(technology);
 		g_clear_error(&error);
 	}
+
+	technology->tethering_ident = g_key_file_get_string(keyfile,
+				identifier, "Tethering.Identifier", NULL);
+
+	technology->tethering_passphrase = g_key_file_get_string(keyfile,
+				identifier, "Tethering.Passphrase", NULL);
 done:
 	g_free(identifier);
 
@@ -510,7 +547,8 @@ void __connman_technology_list_struct(DBusMessageIter *array)
 	for (list = technology_list; list; list = list->next) {
 		struct connman_technology *technology = list->data;
 
-		if (technology->path == NULL)
+		if (technology->path == NULL ||
+					technology->hardblocked == TRUE)
 			continue;
 
 		dbus_message_iter_open_container(array, DBUS_TYPE_STRUCT,
@@ -542,12 +580,10 @@ static gboolean technology_pending_reply(gpointer user_data)
 }
 
 static int technology_enable(struct connman_technology *technology,
-		DBusMessage *msg)
+						connman_bool_t hardblock)
 {
 	GSList *list;
 	int err = 0;
-	int ret = -ENODEV;
-	DBusMessage *reply;
 
 	DBG("technology %p enable", technology);
 
@@ -562,69 +598,26 @@ static int technology_enable(struct connman_technology *technology,
 		goto done;
 	}
 
-	if (msg != NULL) {
-		/*
-		 * This is a bit of a trick. When msg is not NULL it means
-		 * thats technology_enable was invoked from the manager API.
-		 * Hence we save the state here.
-		 */
-		technology->enable_persistent = TRUE;
-		save_state(technology);
-	}
+	if (hardblock == TRUE && technology->enable_persistent == FALSE)
+		goto done;
 
 	__connman_rfkill_block(technology->type, FALSE);
 
-	/*
-	 * An empty device list means that devices in the technology
-	 * were rfkill blocked. The unblock above will enable the devs.
-	 */
-	if (technology->device_list == NULL) {
-		ret = 0;
-		goto done;
-	}
-
 	for (list = technology->device_list; list; list = list->next) {
 		struct connman_device *device = list->data;
 
 		err = __connman_device_enable(device);
-		/*
-		 * err = 0 : Device was enabled right away.
-		 * If atleast one device gets enabled, we consider
-		 * the technology to be enabled.
-		 */
-		if (err == 0)
-			ret = 0;
 	}
 
 done:
-	if (ret == 0) {
-		if (msg != NULL)
-			g_dbus_send_reply(connection, msg, DBUS_TYPE_INVALID);
-		return ret;
-	}
-
-	if (msg != NULL) {
-		if (err == -EINPROGRESS) {
-			technology->pending_reply = dbus_message_ref(msg);
-			technology->pending_timeout = g_timeout_add_seconds(10,
-					technology_pending_reply, technology);
-		} else {
-			reply = __connman_error_failed(msg, -err);
-			if (reply != NULL)
-				g_dbus_send_message(connection, reply);
-		}
-	}
-
 	return err;
 }
 
 static int technology_disable(struct connman_technology *technology,
-		DBusMessage *msg)
+						connman_bool_t hardblock)
 {
 	GSList *list;
 	int err = 0;
-	int ret = -ENODEV;
-	DBusMessage *reply;
 
 	DBG("technology %p disable", technology);
 
@@ -642,43 +635,58 @@ static int technology_disable(struct connman_technology *technology,
 	if (technology->tethering == TRUE)
 		set_tethering(technology, FALSE);
 
-	if (msg != NULL) {
-		technology->enable_persistent = FALSE;
-		save_state(technology);
-	}
-
-	__connman_rfkill_block(technology->type, TRUE);
+	if (hardblock == FALSE)
+		__connman_rfkill_block(technology->type, TRUE);
 
 	for (list = technology->device_list; list; list = list->next) {
 		struct connman_device *device = list->data;
 
 		err = __connman_device_disable(device);
-		if (err == 0)
-			ret = 0;
 	}
 
 done:
-	if (ret == 0) {
-		if (msg != NULL)
-			g_dbus_send_reply(connection, msg, DBUS_TYPE_INVALID);
-		return ret;
-	}
-
-	if (msg != NULL) {
-		if (err == -EINPROGRESS) {
-			technology->pending_reply = dbus_message_ref(msg);
-			technology->pending_timeout = g_timeout_add_seconds(10,
-					technology_pending_reply, technology);
-		} else {
-			reply = __connman_error_failed(msg, -err);
-			if (reply != NULL)
-				g_dbus_send_message(connection, reply);
-		}
-	}
-
 	return err;
 }
 
+static DBusMessage *set_powered(struct connman_technology *technology,
+				DBusMessage *msg, connman_bool_t powered)
+{
+	DBusMessage *reply = NULL;
+	int err = 0;
+
+	if (technology->hardblocked == TRUE) {
+		err = -EACCES;
+		goto make_reply;
+	}
+
+	if (powered == TRUE)
+		err = technology_enable(technology, FALSE);
+	else
+		err = technology_disable(technology, FALSE);
+
+	if (err != -EBUSY) {
+		technology->enable_persistent = powered;
+		technology_save(technology);
+	}
+
+make_reply:
+	if (err == -EINPROGRESS) {
+		technology->pending_reply = dbus_message_ref(msg);
+		technology->pending_timeout = g_timeout_add_seconds(10,
+					technology_pending_reply, technology);
+	} else if (err == -EALREADY) {
+		if (powered == TRUE)
+			reply = __connman_error_already_enabled(msg);
+		else
+			reply = __connman_error_already_disabled(msg);
+	} else if (err < 0)
+		reply = __connman_error_failed(msg, -err);
+	else
+		reply = g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+
+	return reply;
+}
+
 static DBusMessage *set_property(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -716,8 +724,12 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 		dbus_message_iter_get_basic(&value, &tethering);
 
-		if (technology->tethering == tethering)
-			return __connman_error_in_progress(msg);
+		if (technology->tethering == tethering) {
+			if (tethering == FALSE)
+				return __connman_error_already_disabled(msg);
+			else
+				return __connman_error_already_enabled(msg);
+		}
 
 		err = set_tethering(technology, tethering);
 		if (err < 0)
@@ -731,7 +743,20 @@ static DBusMessage *set_property(DBusConnection *conn,
 		if (technology->type != CONNMAN_SERVICE_TYPE_WIFI)
 			return __connman_error_not_supported(msg);
 
-		technology->tethering_ident = g_strdup(str);
+		if (strlen(str) < 1 || strlen(str) > 32)
+			return __connman_error_invalid_arguments(msg);
+
+		if (g_strcmp0(technology->tethering_ident, str) != 0) {
+			g_free(technology->tethering_ident);
+			technology->tethering_ident = g_strdup(str);
+			technology_save(technology);
+
+			connman_dbus_property_changed_basic(technology->path,
+						CONNMAN_TECHNOLOGY_INTERFACE,
+						"TetheringIdentifier",
+						DBUS_TYPE_STRING,
+						&technology->tethering_ident);
+		}
 	} else if (g_str_equal(name, "TetheringPassphrase") == TRUE) {
 		const char *str;
 
@@ -740,10 +765,20 @@ static DBusMessage *set_property(DBusConnection *conn,
 		if (technology->type != CONNMAN_SERVICE_TYPE_WIFI)
 			return __connman_error_not_supported(msg);
 
-		if (strlen(str) < 8)
-			return __connman_error_invalid_arguments(msg);
+		if (strlen(str) < 8 || strlen(str) > 63)
+			return __connman_error_passphrase_required(msg);
 
-		technology->tethering_passphrase = g_strdup(str);
+		if (g_strcmp0(technology->tethering_passphrase, str) != 0) {
+			g_free(technology->tethering_passphrase);
+			technology->tethering_passphrase = g_strdup(str);
+			technology_save(technology);
+
+			connman_dbus_property_changed_basic(technology->path,
+					CONNMAN_TECHNOLOGY_INTERFACE,
+					"TetheringPassphrase",
+					DBUS_TYPE_STRING,
+					&technology->tethering_passphrase);
+		}
 	} else if (g_str_equal(name, "Powered") == TRUE) {
 		connman_bool_t enable;
 
@@ -751,11 +786,8 @@ static DBusMessage *set_property(DBusConnection *conn,
 			return __connman_error_invalid_arguments(msg);
 
 		dbus_message_iter_get_basic(&value, &enable);
-		if (enable == TRUE)
-			technology_enable(technology, msg);
-		else
-			technology_disable(technology, msg);
 
+		return set_powered(technology, msg, enable);
 	} else
 		return __connman_error_invalid_property(msg);
 
@@ -831,7 +863,7 @@ void __connman_technology_scan_stopped(struct connman_device *device)
 		if (__connman_device_get_service_type(other_device) != type)
 			continue;
 
-		if (__connman_device_scanning(other_device))
+		if (connman_device_get_scanning(other_device) == TRUE)
 			count += 1;
 	}
 
@@ -839,6 +871,31 @@ void __connman_technology_scan_stopped(struct connman_device *device)
 		reply_scan_pending(technology, 0);
 }
 
+void __connman_technology_notify_regdom_by_device(struct connman_device *device,
+						int result, const char *alpha2)
+{
+	struct connman_technology *technology;
+	enum connman_service_type type;
+
+	type = __connman_device_get_service_type(device);
+	technology = technology_find(type);
+
+	if (technology == NULL)
+		return;
+
+	if (result < 0) {
+		if (technology->driver != NULL &&
+				technology->driver->set_regdom != NULL) {
+			technology->driver->set_regdom(technology, alpha2);
+			return;
+		}
+
+		alpha2 = NULL;
+	}
+
+	connman_technology_regdom_notify(technology, alpha2);
+}
+
 static DBusMessage *scan(DBusConnection *conn, DBusMessage *msg, void *data)
 {
 	struct connman_technology *technology = data;
@@ -858,19 +915,43 @@ static DBusMessage *scan(DBusConnection *conn, DBusMessage *msg, void *data)
 	return NULL;
 }
 
-static GDBusMethodTable technology_methods[] = {
-	{ "GetProperties", "",   "a{sv}", get_properties },
-	{ "SetProperty",   "sv", "",      set_property   },
-	{ "Scan",          "",    "",     scan,
-						G_DBUS_METHOD_FLAG_ASYNC },
+static const GDBusMethodTable technology_methods[] = {
+	{ GDBUS_DEPRECATED_METHOD("GetProperties",
+			NULL, GDBUS_ARGS({ "properties", "a{sv}" }),
+			get_properties) },
+	{ GDBUS_ASYNC_METHOD("SetProperty",
+			GDBUS_ARGS({ "name", "s" }, { "value", "v" }),
+			NULL, set_property) },
+	{ GDBUS_ASYNC_METHOD("Scan", NULL, NULL, scan) },
 	{ },
 };
 
-static GDBusSignalTable technology_signals[] = {
-	{ "PropertyChanged", "sv" },
+static const GDBusSignalTable technology_signals[] = {
+	{ GDBUS_SIGNAL("PropertyChanged",
+			GDBUS_ARGS({ "name", "s" }, { "value", "v" })) },
 	{ },
 };
 
+static gboolean technology_dbus_register(struct connman_technology *technology)
+{
+	if (technology->dbus_registered == TRUE ||
+					technology->hardblocked == TRUE)
+		return TRUE;
+
+	if (g_dbus_register_interface(connection, technology->path,
+				CONNMAN_TECHNOLOGY_INTERFACE,
+				technology_methods, technology_signals,
+				NULL, technology, NULL) == FALSE) {
+		connman_error("Failed to register %s", technology->path);
+		return FALSE;
+	}
+
+	technology_added_signal(technology);
+	technology->dbus_registered = TRUE;
+
+	return TRUE;
+}
+
 static struct connman_technology *technology_get(enum connman_service_type type)
 {
 	struct connman_technology *technology;
@@ -913,6 +994,11 @@ static struct connman_technology *technology_get(enum connman_service_type type)
 
 	technology->refcount = 1;
 
+	if (type == CONNMAN_SERVICE_TYPE_ETHERNET)
+		technology->hardblocked = FALSE;
+	else
+		technology->hardblocked = TRUE;
+
 	technology->type = type;
 	technology->path = g_strdup_printf("%s/technology/%s",
 							CONNMAN_PATH, str);
@@ -921,20 +1007,14 @@ static struct connman_technology *technology_get(enum connman_service_type type)
 
 	technology->pending_reply = NULL;
 
-	load_state(technology);
+	technology_load(technology);
 
-	if (g_dbus_register_interface(connection, technology->path,
-					CONNMAN_TECHNOLOGY_INTERFACE,
-					technology_methods, technology_signals,
-					NULL, technology, NULL) == FALSE) {
-		connman_error("Failed to register %s", technology->path);
+	if (technology_dbus_register(technology) == FALSE) {
 		g_free(technology);
 		return NULL;
 	}
 
-	technology_list = g_slist_append(technology_list, technology);
-
-	technology_added_signal(technology);
+	technology_list = g_slist_prepend(technology_list, technology);
 
 	technology->driver = driver;
 	err = driver->probe(technology);
@@ -946,6 +1026,18 @@ static struct connman_technology *technology_get(enum connman_service_type type)
 	return technology;
 }
 
+static void technology_dbus_unregister(struct connman_technology *technology)
+{
+	if (technology->dbus_registered == FALSE)
+		return;
+
+	technology_removed_signal(technology);
+	g_dbus_unregister_interface(connection, technology->path,
+		CONNMAN_TECHNOLOGY_INTERFACE);
+
+	technology->dbus_registered = FALSE;
+}
+
 static void technology_put(struct connman_technology *technology)
 {
 	DBG("technology %p", technology);
@@ -962,15 +1054,14 @@ static void technology_put(struct connman_technology *technology)
 
 	technology_list = g_slist_remove(technology_list, technology);
 
-	technology_removed_signal(technology);
-
-	g_dbus_unregister_interface(connection, technology->path,
-						CONNMAN_TECHNOLOGY_INTERFACE);
+	technology_dbus_unregister(technology);
 
 	g_slist_free(technology->device_list);
 
 	g_free(technology->path);
 	g_free(technology->regdom);
+	g_free(technology->tethering_ident);
+	g_free(technology->tethering_passphrase);
 	g_free(technology);
 }
 
@@ -1060,7 +1151,9 @@ int __connman_technology_add_device(struct connman_device *device)
 		return -ENXIO;
 	}
 
-	if (technology->enable_persistent && !global_offlinemode) {
+	if (technology->enable_persistent &&
+					global_offlinemode == FALSE &&
+					technology->hardblocked == FALSE) {
 		int err = __connman_device_enable(device);
 		/*
 		 * connman_technology_add_device() calls __connman_device_enable()
@@ -1071,11 +1164,12 @@ int __connman_technology_add_device(struct connman_device *device)
 		if (err == -EALREADY)
 			__connman_technology_enabled(type);
 	}
-	/* if technology persistent state is offline */
-	if (!technology->enable_persistent)
+	/* if technology persistent state is offline or hardblocked */
+	if (technology->enable_persistent == FALSE ||
+					technology->hardblocked == TRUE)
 		__connman_device_disable(device);
 
-	technology->device_list = g_slist_append(technology->device_list,
+	technology->device_list = g_slist_prepend(technology->device_list,
 								device);
 
 	return 0;
@@ -1097,9 +1191,6 @@ int __connman_technology_remove_device(struct connman_device *device)
 		return -ENXIO;
 	}
 
-	if (__connman_device_scanning(device))
-		__connman_technology_scan_stopped(device);
-
 	technology->device_list = g_slist_remove(technology->device_list,
 								device);
 	technology_put(technology);
@@ -1130,8 +1221,10 @@ int __connman_technology_enabled(enum connman_service_type type)
 	if (technology == NULL)
 		return -ENXIO;
 
-	if (__sync_fetch_and_add(&technology->enabled, 1) == 0)
-		powered_changed(technology);
+	if (__sync_fetch_and_add(&technology->enabled, 1) != 0)
+		return -EALREADY;
+
+	powered_changed(technology);
 
 	if (technology->pending_reply != NULL) {
 		g_dbus_send_reply(connection, technology->pending_reply, DBUS_TYPE_INVALID);
@@ -1152,6 +1245,9 @@ int __connman_technology_disabled(enum connman_service_type type)
 	if (technology == NULL)
 		return -ENXIO;
 
+	if (__sync_fetch_and_sub(&technology->enabled, 1) != 1)
+		return -EINPROGRESS;
+
 	if (technology->pending_reply != NULL) {
 		g_dbus_send_reply(connection, technology->pending_reply, DBUS_TYPE_INVALID);
 		dbus_message_unref(technology->pending_reply);
@@ -1160,8 +1256,7 @@ int __connman_technology_disabled(enum connman_service_type type)
 		technology->pending_timeout = 0;
 	}
 
-	if (__sync_fetch_and_sub(&technology->enabled, 1) == 1)
-		powered_changed(technology);
+	powered_changed(technology);
 
 	return 0;
 }
@@ -1192,10 +1287,10 @@ int __connman_technology_set_offlinemode(connman_bool_t offlinemode)
 		struct connman_technology *technology = list->data;
 
 		if (offlinemode)
-			err = technology_disable(technology, NULL);
+			err = technology_disable(technology, FALSE);
 
 		if (!offlinemode && technology->enable_persistent)
-			err = technology_enable(technology, NULL);
+			err = technology_enable(technology, FALSE);
 	}
 
 	if (err == 0 || err == -EINPROGRESS || err == -EALREADY) {
@@ -1225,6 +1320,43 @@ void __connman_technology_set_connected(enum connman_service_type type,
 			DBUS_TYPE_BOOLEAN, &connected);
 }
 
+static void technology_apply_hardblock_change(struct connman_technology *technology,
+						connman_bool_t hardblock)
+{
+	gboolean apply = TRUE;
+	GList *start, *list;
+
+	if (technology->hardblocked == hardblock)
+		return;
+
+	start = g_hash_table_get_values(rfkill_list);
+	for (list = start; list != NULL; list = list->next) {
+		struct connman_rfkill *rfkill = list->data;
+
+		if (rfkill->type != technology->type)
+			continue;
+
+		if (rfkill->hardblock != hardblock)
+			apply = FALSE;
+	}
+
+	g_list_free(start);
+
+	if (apply == FALSE)
+		return;
+
+	technology->hardblocked = hardblock;
+
+	if (hardblock == TRUE) {
+		DBG("%s is switched off.", get_name(technology->type));
+		technology_disable(technology, TRUE);
+		technology_dbus_unregister(technology);
+	} else {
+		technology_enable(technology, TRUE);
+		technology_dbus_register(technology);
+	}
+}
+
 int __connman_technology_add_rfkill(unsigned int index,
 					enum connman_service_type type,
 						connman_bool_t softblock,
@@ -1236,7 +1368,7 @@ int __connman_technology_add_rfkill(unsigned int index,
 	DBG("index %u type %d soft %u hard %u", index, type,
 							softblock, hardblock);
 
-	rfkill = g_hash_table_lookup(rfkill_list, &index);
+	rfkill = g_hash_table_lookup(rfkill_list, GINT_TO_POINTER(index));
 	if (rfkill != NULL)
 		goto done;
 
@@ -1249,7 +1381,7 @@ int __connman_technology_add_rfkill(unsigned int index,
 	rfkill->softblock = softblock;
 	rfkill->hardblock = hardblock;
 
-	g_hash_table_insert(rfkill_list, &rfkill->index, rfkill);
+	g_hash_table_insert(rfkill_list, GINT_TO_POINTER(index), rfkill);
 
 done:
 	technology = technology_get(type);
@@ -1257,10 +1389,7 @@ done:
 	if (technology == NULL)
 		return -ENXIO;
 
-	if (hardblock) {
-		DBG("%s is switched off.", get_name(type));
-		return 0;
-	}
+	technology_apply_hardblock_change(technology, hardblock);
 
 	/*
 	 * If Offline mode is on, we softblock the device if it isnt already.
@@ -1290,7 +1419,7 @@ int __connman_technology_update_rfkill(unsigned int index,
 
 	DBG("index %u soft %u hard %u", index, softblock, hardblock);
 
-	rfkill = g_hash_table_lookup(rfkill_list, &index);
+	rfkill = g_hash_table_lookup(rfkill_list, GINT_TO_POINTER(index));
 	if (rfkill == NULL)
 		return -ENXIO;
 
@@ -1301,16 +1430,13 @@ int __connman_technology_update_rfkill(unsigned int index,
 	rfkill->softblock = softblock;
 	rfkill->hardblock = hardblock;
 
-	if (hardblock) {
-		DBG("%s is switched off.", get_name(type));
-		return 0;
-	}
-
 	technology = technology_find(type);
 	/* If there is no driver for this type, ignore it. */
 	if (technology == NULL)
 		return -ENXIO;
 
+	technology_apply_hardblock_change(technology, hardblock);
+
 	if (!global_offlinemode) {
 		if (technology->enable_persistent && softblock)
 			return __connman_rfkill_block(type, FALSE);
@@ -1329,11 +1455,11 @@ int __connman_technology_remove_rfkill(unsigned int index,
 
 	DBG("index %u", index);
 
-	rfkill = g_hash_table_lookup(rfkill_list, &index);
+	rfkill = g_hash_table_lookup(rfkill_list, GINT_TO_POINTER(index));
 	if (rfkill == NULL)
 		return -ENXIO;
 
-	g_hash_table_remove(rfkill_list, &index);
+	g_hash_table_remove(rfkill_list, GINT_TO_POINTER(index));
 
 	technology = technology_find(type);
 	if (technology == NULL)
@@ -1350,7 +1476,7 @@ int __connman_technology_init(void)
 
 	connection = connman_dbus_get_connection();
 
-	rfkill_list = g_hash_table_new_full(g_int_hash, g_int_equal,
+	rfkill_list = g_hash_table_new_full(g_direct_hash, g_direct_equal,
 							NULL, free_rfkill);
 
 	global_offlinemode = connman_technology_load_offlinemode();
