@@ -33,12 +33,6 @@
 #define STATUS_URL_IPV4  "http://ipv4.connman.net/online/status.html"
 #define STATUS_URL_IPV6  "http://ipv6.connman.net/online/status.html"
 
-#define wispr_portal_context_ref(_wp_ctxt) \
-	wispr_portal_context_ref_debug(_wp_ctxt, __FILE__, __LINE__, __func__)
-
-#define wispr_portal_context_unref(_wp_ctxt) \
-	wispr_portal_context_unref_debug(_wp_ctxt, __FILE__, __LINE__, __func__)
-
 struct connman_wispr_message {
 	gboolean has_error;
 	const char *current_element;
@@ -65,10 +59,9 @@ struct wispr_route {
 };
 
 struct connman_wispr_portal_context {
-	int refcount;
-
 	struct connman_service *service;
 	enum connman_ipconfig_type type;
+	struct connman_wispr_portal *wispr_portal;
 
 	/* Portal/WISPr common */
 	GWeb *web;
@@ -167,8 +160,13 @@ static void free_connman_wispr_portal_context(struct connman_wispr_portal_contex
 	if (wp_context == NULL)
 		return;
 
-	if (wp_context->service != NULL)
-		connman_service_unref(wp_context->service);
+	if (wp_context->wispr_portal != NULL) {
+		if (wp_context->wispr_portal->ipv4_context == wp_context)
+			wp_context->wispr_portal->ipv4_context = NULL;
+
+		if (wp_context->wispr_portal->ipv6_context == wp_context)
+			wp_context->wispr_portal->ipv6_context = NULL;
+	}
 
 	if (wp_context->token > 0)
 		connman_proxy_lookup_cancel(wp_context->token);
@@ -195,60 +193,9 @@ static void free_connman_wispr_portal_context(struct connman_wispr_portal_contex
 	g_free(wp_context);
 }
 
-static struct connman_wispr_portal_context *
-wispr_portal_context_ref_debug(struct connman_wispr_portal_context *wp_context,
-				const char *file, int line, const char *caller)
-{
-	DBG("%p ref %d by %s:%d:%s()", wp_context, wp_context->refcount + 1,
-							file, line, caller);
-
-	__sync_fetch_and_add(&wp_context->refcount, 1);
-
-	return wp_context;
-}
-
-static struct connman_wispr_portal_context *
-wispr_portal_context_unref_debug(struct connman_wispr_portal_context *wp_context,
-				const char *file, int line, const char *caller)
-{
-	DBG("%p ref %d by %s:%d:%s()", wp_context, wp_context->refcount - 1,
-							file, line, caller);
-
-	if (__sync_fetch_and_sub(&wp_context->refcount, 1) != 1)
-		return wp_context;
-
-	free_connman_wispr_portal_context(wp_context);
-
-	return NULL;
-}
-
 static struct connman_wispr_portal_context *create_wispr_portal_context(void)
 {
-	struct connman_wispr_portal_context *wp_context = NULL;
-
-	 wp_context = g_try_new0(struct connman_wispr_portal_context, 1);
-	 if (wp_context == NULL)
-		return NULL;
-
-	 wp_context->refcount = 1;
-
-	 return wp_context;
-}
-
-/**
- * This function is usued only by free_connman_wispr_portal, context pointer
- * might still be owned by a third party (gweb, agent dbus call...)
- * so we don't want service to be referenced in any context when it is not
- * valid anymore.
- */
-static void
-reset_service_usage(struct connman_wispr_portal_context *wp_context)
-{
-	if (wp_context->service == NULL)
-		return;
-
-	connman_service_unref(wp_context->service);
-	wp_context->service = NULL;
+	return g_try_new0(struct connman_wispr_portal_context, 1);
 }
 
 static void free_connman_wispr_portal(gpointer data)
@@ -260,15 +207,8 @@ static void free_connman_wispr_portal(gpointer data)
 	if (wispr_portal == NULL)
 		return;
 
-	if (wispr_portal->ipv4_context != NULL) {
-		reset_service_usage(wispr_portal->ipv4_context);
-		wispr_portal_context_unref(wispr_portal->ipv4_context);
-	}
-
-	if (wispr_portal->ipv6_context != NULL) {
-		reset_service_usage(wispr_portal->ipv6_context);
-		wispr_portal_context_unref(wispr_portal->ipv6_context);
-	}
+	free_connman_wispr_portal_context(wispr_portal->ipv4_context);
+	free_connman_wispr_portal_context(wispr_portal->ipv6_context);
 
 	g_free(wispr_portal);
 }
@@ -562,8 +502,6 @@ static void wispr_portal_request_portal(struct connman_wispr_portal_context *wp_
 
 	if (wp_context->request_id == 0)
 		wispr_portal_error(wp_context);
-	else
-		wispr_portal_context_ref(wp_context);
 }
 
 static gboolean wispr_input(const guint8 **data, gsize *length,
@@ -605,7 +543,6 @@ static void wispr_portal_browser_reply_cb(struct connman_service *service,
 
 	DBG("");
 
-	wp_context = wispr_portal_context_unref(wp_context);
 	if (service == NULL || wp_context == NULL)
 		return;
 
@@ -630,17 +567,17 @@ static void wispr_portal_request_wispr_login(struct connman_service *service,
 
 	DBG("");
 
-	wp_context = wispr_portal_context_unref(wp_context);
-	if (wp_context == NULL)
-		return;
-
-	if (error != NULL && g_strcmp0(error,
+	if (error != NULL) {
+		if (g_strcmp0(error,
 			"net.connman.Agent.Error.LaunchBrowser") == 0) {
-		wispr_portal_context_ref(wp_context);
+			if (__connman_agent_request_browser(service,
+					wispr_portal_browser_reply_cb,
+					wp_context->redirect_url,
+					wp_context) != -EINPROGRESS)
+				free_connman_wispr_portal_context(wp_context);
+		} else
+			free_connman_wispr_portal_context(wp_context);
 
-		__connman_agent_request_browser(service,
-				wispr_portal_browser_reply_cb,
-				wp_context->redirect_url, wp_context);
 		return;
 	}
 
@@ -655,8 +592,6 @@ static void wispr_portal_request_wispr_login(struct connman_service *service,
 					"application/x-www-form-urlencoded",
 					wispr_input, wispr_portal_web_result,
 					wp_context);
-	if (wp_context->request_id != 0)
-		wispr_portal_context_ref(wp_context);
 
 	connman_wispr_message_init(&wp_context->wispr_msg);
 }
@@ -694,13 +629,11 @@ static gboolean wispr_manage_message(GWebResult *result,
 
 		wp_context->wispr_result = CONNMAN_WISPR_RESULT_LOGIN;
 
-		wispr_portal_context_ref(wp_context);
-
 		if (__connman_agent_request_login_input(wp_context->service,
 					wispr_portal_request_wispr_login,
 					wp_context) != -EINPROGRESS) {
-			wispr_portal_context_unref(wp_context);
 			wispr_portal_error(wp_context);
+			free_connman_wispr_portal_context(wp_context);
 		}
 
 		break;
@@ -743,19 +676,10 @@ static gboolean wispr_portal_web_result(GWebResult *result, gpointer user_data)
 
 	DBG("");
 
-	wp_context = wispr_portal_context_unref(wp_context);
-	if (wp_context == NULL)
-		return FALSE;
-
-	if (wp_context->request_id == 0)
-		return FALSE;
-
 	if (wp_context->wispr_result != CONNMAN_WISPR_RESULT_ONLINE) {
 		g_web_result_get_chunk(result, &chunk, &length);
 
 		if (length > 0) {
-			wispr_portal_context_ref(wp_context);
-
 			g_web_parser_feed_data(wp_context->wispr_parser,
 								chunk, length);
 			return TRUE;
@@ -779,22 +703,21 @@ static gboolean wispr_portal_web_result(GWebResult *result, gpointer user_data)
 			break;
 
 		if (g_web_result_get_header(result, "X-ConnMan-Status",
-								&str) == TRUE)
+						&str) == TRUE) {
 			portal_manage_status(result, wp_context);
-		else {
-			wispr_portal_context_ref(wp_context);
-
+			free_connman_wispr_portal_context(wp_context);
+			return FALSE;
+		}
+		else
 			__connman_agent_request_browser(wp_context->service,
 					wispr_portal_browser_reply_cb,
 					wp_context->redirect_url, wp_context);
-		}
 
 		break;
 	case 302:
 		if (g_web_supports_tls() == FALSE ||
 				g_web_result_get_header(result, "Location",
 							&redirect) == FALSE) {
-			wispr_portal_context_ref(wp_context);
 
 			__connman_agent_request_browser(wp_context->service,
 					wispr_portal_browser_reply_cb,
@@ -809,15 +732,16 @@ static gboolean wispr_portal_web_result(GWebResult *result, gpointer user_data)
 		wp_context->request_id = g_web_request_get(wp_context->web,
 				redirect, wispr_portal_web_result,
 				wispr_route_request, wp_context);
-		if (wp_context->request_id != 0)
-			wispr_portal_context_ref(wp_context);
 
 		goto done;
 	case 400:
 	case 404:
 		if (__connman_service_online_check_failed(wp_context->service,
-							wp_context->type) == 0)
+						wp_context->type) == 0) {
 			wispr_portal_error(wp_context);
+			free_connman_wispr_portal_context(wp_context);
+			return FALSE;
+		}
 
 		break;
 	default:
@@ -837,7 +761,6 @@ static void proxy_callback(const char *proxy, void *user_data)
 
 	DBG("proxy %s", proxy);
 
-	wp_context = wispr_portal_context_unref(wp_context);
 	if (wp_context == NULL)
 		return;
 
@@ -950,13 +873,12 @@ static int wispr_portal_detect(struct connman_wispr_portal_context *wp_context)
 
 		if (wp_context->token == 0) {
 			err = -EINVAL;
-			goto done;
+			free_connman_wispr_portal_context(wp_context);
 		}
 	} else {
 		g_timeout_add_seconds(0, no_proxy_callback, wp_context);
 	}
 
-	wispr_portal_context_ref(wp_context);
 done:
 	g_strfreev(nameservers);
 
@@ -1000,16 +922,15 @@ int __connman_wispr_start(struct connman_service *service,
 
 	/* If there is already an existing context, we wipe it */
 	if (wp_context != NULL)
-		wispr_portal_context_unref(wp_context);
+		free_connman_wispr_portal_context(wp_context);
 
 	wp_context = create_wispr_portal_context();
 	if (wp_context == NULL)
 		return -ENOMEM;
 
-	connman_service_ref(service);
-
 	wp_context->service = service;
 	wp_context->type = type;
+	wp_context->wispr_portal = wispr_portal;
 
 	if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
 		wispr_portal->ipv4_context = wp_context;
